/********************************************************************
* FILENAME: inst.cpp
*
* DESCRIPTION:
*	NT API interface DLL used by Installer
*
* REFERENCES:
*
*	inst.h
*
	Much of the security source code here was obtained from Michael Nelson.
	I am on his 'drop list' for source code enhancements to the
	DCOMPERM com object he has developed.
*

********************************************************************/

#define	_WIN32_DCOM
#include <windows.h>
#include <lm.h>
#include <comcat.h>
#include <comdef.h>
#include <iostream.h>
#include <winsvc.h>
#include "ntsecapi.h"

#define REPORT_CALL_FAILURES
#include "utils.h"

//	compiler interface
#include <tchar.h>
#include <stdio.h>
#include <string.h>

#include "inst.h"

BOOL WINAPI DllMain( PVOID hmod, ULONG ulReason, PCONTEXT pctx )
{
        return TRUE;
}

inline
const TCHAR *FormatLastError(DWORD lastError, TCHAR *formatBuffer, const TCHAR *caption)
{
	TCHAR errorText[4096];

	if (!FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM, 0, lastError, 0, errorText, sizeof(errorText), 0))
		_tcscpy(errorText, __T("Unknown error"));

	_tcstok(errorText, __T("\n\r"));	//	null terminate
	_stprintf(formatBuffer, __T("%s, \"%s\" (0x%x)\n"), caption, errorText, lastError);

	return formatBuffer;
}

void ShowLastError()
{
	char *msgBuf;

	FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,    NULL,
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &msgBuf,    0,    NULL );// Display the string.

	MessageBox( NULL, msgBuf, "GetLastError", MB_OK|MB_ICONINFORMATION );
    LocalFree(msgBuf);
}

extern "C" int __declspec(dllexport) AddTCBPrivilege()
{
	_bstr_t wacct( "Administrators" );
    LSA_HANDLE PolicyHandle;
    PSID pSid = 0;
    NTSTATUS Status;
    int iRetVal=0;

    Status = OpenPolicy( POLICY_CREATE_ACCOUNT | POLICY_LOOKUP_NAMES, &PolicyHandle );

    if(Status != STATUS_SUCCESS)
        return 0;

	// note: wacct is in format of computername\username
    if(GetPrincipalSID( wacct, &pSid ) == ERROR_SUCCESS)
		{
        // Grant the SeServiceLogonRight to users represented by pSid.
        Status = SetPrivilegeOnAccount( PolicyHandle, pSid, L"SeTcbPrivilege" );

        if(Status == STATUS_SUCCESS)
            iRetVal = 1;
        else
            iRetVal = 0;
		}
    else
		iRetVal = 0;

    // Close the policy handle.
    LsaClose(PolicyHandle);

    // Free memory allocated for SID.
    if(pSid != NULL) HeapFree(GetProcessHeap(), 0, pSid);

    return iRetVal;

}

extern "C" int __declspec(dllexport) VerifyTCBPrivilege()
{
	HANDLE procToken;
	int ret = 1;

	if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES|TOKEN_QUERY, &procToken))
		{
		TOKEN_PRIVILEGES tp;

		tp.PrivilegeCount = 1;
		tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
		
		if ( LookupPrivilegeValue(NULL, SE_TCB_NAME, &tp.Privileges[0].Luid) )
			if ( AdjustTokenPrivileges(procToken, FALSE, &tp, sizeof(tp), NULL, NULL) )
				if ( GetLastError() == ERROR_NOT_ALL_ASSIGNED )
					ret = 0;
		}
	else
	{
#ifndef	NDEBUG
		char formatBuffer[4096];
		FormatLastError(GetLastError(), formatBuffer, "OpenProcessToken()");
		MessageBox(0, formatBuffer, "ERROR!!!!!", MB_OK);
#endif

		return -1;
	}

	CloseHandle( procToken );

	return ret;
}

extern "C" int __declspec(dllexport) VerifyRemoteServer( char * ntMachineName, char * user, char * password, char * version, char * revision, char * installed, char *ipHostName )
{
	HANDLE	token, procToken;
	HKEY	key,rkey;
	DWORD	disposition;
	LONG	error;
	DWORD	lType, size;
	char	keyToFind[128];

	if (OpenProcessToken(GetCurrentProcess(), TOKEN_ADJUST_PRIVILEGES|TOKEN_QUERY, &procToken))
		{
		TOKEN_PRIVILEGES tp;

		tp.PrivilegeCount = 1;
		tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;
		LookupPrivilegeValue(NULL, SE_TCB_NAME, &tp.Privileges[0].Luid);
		if ( !AdjustTokenPrivileges(procToken, FALSE, &tp, sizeof(tp), NULL, NULL) )
			{
			CloseHandle(procToken);
			return 0;
			}
		}
	else
		return 0;


	if ( !LogonUser( user, ".", password, LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_DEFAULT, &token ) )
		{
		return -1;
		}

	if ( !ImpersonateLoggedOnUser( token ) )
		{
		return 0;
		}

	if ( RegConnectRegistry( ntMachineName, HKEY_LOCAL_MACHINE, &key ) != ERROR_SUCCESS )
		{
		// cleanup
		RevertToSelf();
		CloseHandle( token );
		return -4;
		}

	sprintf ( keyToFind, "SOFTWARE\\MyCompany\\MySoftware\\%s", version );
	error = RegCreateKeyEx(key, keyToFind,	
		0, "", REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS, NULL, &rkey, &disposition);
	if(!rkey) 
		{
		// talxware not installed or wrong version
		// cleanup
		RegCloseKey( key );
		RevertToSelf();
		CloseHandle( token );
		return -2;
		}


	size = 64;
	error = RegQueryValueEx(rkey, "Installed", NULL, &lType, (LPBYTE)installed, &size);
	if(error != ERROR_SUCCESS) 
		{
		// talxware not installed or wrong version
		// cleanup
		RegCloseKey( key );
		RevertToSelf();
		CloseHandle( token );
		return -2;
		} 

	size = 64;
	error = RegQueryValueEx(rkey, "Revision", NULL, &lType, (LPBYTE)revision, &size);
	if(error != ERROR_SUCCESS) 
		{
		// couldn't find revision??
		// cleanup
		RegCloseKey( key );
		RevertToSelf();
		CloseHandle( token );
		return -3;
		}



	size = 64;
	ipHostName[0] = 0;
	HKEY zkey;

	error = RegCreateKeyEx(key, "SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters",	
		0, "", REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS, NULL, &zkey, &disposition);

	error = RegQueryValueEx(zkey, "Hostname", NULL, &lType, (LPBYTE)ipHostName, &size);
	if(error != ERROR_SUCCESS) 
		{
		// couldn't find HostName??
		// cleanup
		RegCloseKey( key );
		RevertToSelf();
		CloseHandle( token );
		return -9;
		} 

	// clean up
	RegCloseKey( key );
	RegCloseKey( rkey );
	RegCloseKey( zkey );
	RevertToSelf();
	CloseHandle( token );

	return 1;
}

extern "C" int __declspec(dllexport) RegisterFile(const char *filename)
{
	int status = 0;

	CoInitialize(0);

	char *entrypoint;
	bool registerserver;
	HINSTANCE hInstance;

	entrypoint = "DllRegisterServer";
	registerserver = true;

	if (HasEXEExtension(filename))		//	executable
	{
		char systemBuffer[1024];

		sprintf(systemBuffer, "%s %s",
			filename,
			registerserver? "/RegServer" : "/UnregServer");

		BOOL retVal;
		STARTUPINFO startupInfo;
		PROCESS_INFORMATION processInfo;

		GetStartupInfo(&startupInfo);
			
		retVal = CreateProcess(
			0,
			systemBuffer, // pointer to command line string
			0,		// pointer to process security attributes
			0,		// pointer to thread security attributes
			FALSE,	// handle inheritance flag
			0,		// creation flags
			0,		// pointer to new environment block
			".",		// pointer to current directory name
			&startupInfo,	// pointer to STARTUPINFO
			&processInfo);	// pointer to PROCESS_INFORMATION);

		if (!retVal)
			status = 1;
		else
		{
			WaitForSingleObject(processInfo.hProcess, INFINITE);
			CloseHandle(processInfo.hThread);
			CloseHandle(processInfo.hProcess);
		}
	}
	else		//	dll
	{
		hInstance = LoadLibrary(filename);
		if (hInstance)
		{
			FARPROC farproc;

			farproc = GetProcAddress(hInstance, entrypoint);
			if (farproc)
			{
				HRESULT hr = farproc();
				if (FAILED(hr))
					status = 1;
			}
			else
				status = 1;

			if (!FreeLibrary(hInstance))
				status = 1;
		}
		else
			status = 1;
	}

 	CoUninitialize();

	return status;
}

extern "C" int __declspec(dllexport) UnRegisterFile(const char *filename)
{
	int status = 0;

	CoInitialize(0);

	char *entrypoint;
	bool registerserver;
	HINSTANCE hInstance;

	entrypoint = "DllUnregisterServer";
	registerserver = false;

	if (HasEXEExtension(filename))		//	executable
	{
		char systemBuffer[1024];

		sprintf(systemBuffer, "%s %s",
			filename,
			registerserver? "/RegServer" : "/UnregServer");

		BOOL retVal;
		STARTUPINFO startupInfo;
		PROCESS_INFORMATION processInfo;

		GetStartupInfo(&startupInfo);
			
		retVal = CreateProcess(
			0,
			systemBuffer, // pointer to command line string
			0,		// pointer to process security attributes
			0,		// pointer to thread security attributes
			FALSE,	// handle inheritance flag
			0,		// creation flags
			0,		// pointer to new environment block
			".",		// pointer to current directory name
			&startupInfo,	// pointer to STARTUPINFO
			&processInfo);	// pointer to PROCESS_INFORMATION);

		if (!retVal)
			status = 1;
		else
		{
			WaitForSingleObject(processInfo.hProcess, INFINITE);
			CloseHandle(processInfo.hThread);
			CloseHandle(processInfo.hProcess);
		}
	}
	else		//	dll
	{
		hInstance = LoadLibrary(filename);
		if (hInstance)
		{
			FARPROC farproc;

			farproc = GetProcAddress(hInstance, entrypoint);
			if (farproc)
			{
				HRESULT hr = farproc();
				if (FAILED(hr))
					status = 1;
			}
			else
				status = 1;

			if (!FreeLibrary(hInstance))
				status = 1;
		}
		else
			status = 1;
	}

 	CoUninitialize();

	return status;
}

extern "C" int __declspec(dllexport) AddUserAccount(const char *acct, const char *pass)
{
	USER_INFO_1 userInfo;
	LOCALGROUP_MEMBERS_INFO_3 lgrpMembers;
	_bstr_t wpass( pass ), wacct( acct );
	NET_API_STATUS status;
	DWORD indx;

	memset( &userInfo, 0, sizeof( userInfo ) );

	userInfo.usri1_name = wacct;
	userInfo.usri1_password = wpass;
	userInfo.usri1_comment = L"!!ADD AN ACCOUNT DESCRIPTION HERE!!";
	userInfo.usri1_priv = USER_PRIV_USER;
	userInfo.usri1_flags = UF_DONT_EXPIRE_PASSWD | UF_SCRIPT | UF_NORMAL_ACCOUNT;

	status = NetUserAdd( NULL, 1, (LPBYTE)&userInfo, &indx );

	if ( status != NERR_Success )
		return status;

	lgrpMembers.lgrmi3_domainandname = wacct;
	status = NetLocalGroupAddMembers( NULL, L"Administrators", 3, (LPBYTE)&lgrpMembers , 1 );

	return status;

}

extern "C" int __declspec(dllexport) RemoveUserAccount(const char *acct)
{
	_bstr_t wacct( acct );
	NET_API_STATUS status;

	status = NetUserDel( NULL, wacct );

	if ( status != NERR_Success )
		return status;

	return status;

}

//	retrieve the machine name
extern "C" int __declspec(dllexport) GetMachineName(char *machineName)
{
	NET_API_STATUS status;
	LPBYTE buffer = NULL;

	status = NetServerGetInfo(NULL, 102, &buffer);

	if (status == NERR_Success) {
		SERVER_INFO_102 *serverInfo = (SERVER_INFO_102 *)buffer;
		size_t machineNameLen = wcslen((wchar_t *)serverInfo->sv102_name);

		wcstombs(machineName, (wchar_t *)serverInfo->sv102_name, machineNameLen);
		machineName[machineNameLen] = 0;

		NetApiBufferFree(buffer);

		return 0;
	} else
		return -1;
}

//	returns 1 if this machine is member of a workgroup, 0 otherwise, -1 failure
extern "C" int __declspec(dllexport) IsWorkgroup()
{
	char machineName[MAX_COMPUTERNAME_LENGTH + 1];

	if (GetMachineName(machineName) == -1)
		return -1;

	//	open policy
	LSA_HANDLE lsaHandle;
	LSA_UNICODE_STRING lsaMachineName;
	size_t machineNameLen = strlen(machineName);
	wchar_t *machineNameW = (wchar_t *)alloca((machineNameLen + 1) * sizeof(wchar_t));
	mbstowcs(machineNameW, machineName, machineNameLen);
	machineNameW[machineNameLen] = 0;

	LSA_OBJECT_ATTRIBUTES lsaObjectAttributes;
	DWORD lsaAccessMask;

	lsaMachineName.Length = machineNameLen * sizeof(wchar_t);
	lsaMachineName.MaximumLength = (machineNameLen + 1) * sizeof(wchar_t);
	lsaMachineName.Buffer = machineNameW;

	ZeroMemory(&lsaObjectAttributes, sizeof(lsaObjectAttributes));
	lsaAccessMask = POLICY_VIEW_LOCAL_INFORMATION;

	if (LsaOpenPolicy(&lsaMachineName, &lsaObjectAttributes, lsaAccessMask, &lsaHandle))
		return -1;

	POLICY_PRIMARY_DOMAIN_INFO *ppdi = 0;

	if(LsaQueryInformationPolicy(lsaHandle, PolicyPrimaryDomainInformation, (void **)&ppdi))
		return -1;

	//	if the SID pointer is null, then it is a WG, some MS sample code
	int isWorkgroup = ppdi->Sid? 0 : 1;

	LsaFreeMemory(ppdi);

	LsaClose(lsaHandle);

	return isWorkgroup;
}

//	returns 1 if this machine is a PDC, 0 otherwise, -1 failure
extern "C" int __declspec(dllexport) IsPrimaryDomainController()
{
	NET_API_STATUS status;
	LPBYTE buffer = NULL;

	status = NetServerGetInfo(NULL, 101, &buffer);

	if (status == NERR_Success) {
		SERVER_INFO_101 *serverInfo = (SERVER_INFO_101 *)buffer;
		int isPDC = serverInfo->sv101_type & SV_TYPE_DOMAIN_CTRL? 1: 0;

		NetApiBufferFree(buffer);
		
		return isPDC;
	}
	else
		return -1;
}

//	returns 1 if this machine is a BDC, 0 otherwise, -1 failure
extern "C" int __declspec(dllexport) IsBackupDomainController()
{
	NET_API_STATUS status;
	LPBYTE buffer = NULL;

	status = NetServerGetInfo(NULL, 101, &buffer);

	if (status == NERR_Success) {
		SERVER_INFO_101 *serverInfo = (SERVER_INFO_101 *)buffer;
		int isBDC = serverInfo->sv101_type & SV_TYPE_DOMAIN_BAKCTRL? 1: 0;

		NetApiBufferFree(buffer);
		
		return isBDC;
	}
	else
		return -1;
}

//	retrieve the domain name
extern "C" int __declspec(dllexport) GetDomainName(const char *machineName, char *domainName)
{
	size_t machineNameLen = strlen(machineName);
	wchar_t *machineNameW = (wchar_t *)alloca((machineNameLen + 1) * sizeof(wchar_t));
	NET_API_STATUS status;
	LPBYTE buffer = NULL;

	mbstowcs(machineNameW, machineName, machineNameLen);
	machineNameW[machineNameLen] = 0;

	status = NetServerGetInfo((char *)machineNameW, 503, &buffer);

	if (status == NERR_Success) {
		SERVER_INFO_503 *serverInfo = (SERVER_INFO_503 *)buffer;
		size_t domainNameLen = wcslen((wchar_t *)serverInfo->sv503_domain);

		wcstombs(domainName, (wchar_t *)serverInfo->sv503_domain, domainNameLen);
		domainName[domainNameLen] = 0;

		NetApiBufferFree(buffer);

		return 0;
	} else
		return -1;
}

//	retrieve the logon server
extern "C" int __declspec(dllexport) GetPDCMachineName(const char *machineName, const char *domainName, char *pdcMachineName)
{
	NET_API_STATUS status;
	LPBYTE buffer = NULL;
	size_t machineNameLen = 0, domainNameLen = 0;
	wchar_t *machineNameW = 0, *domainNameW = 0;

	if (machineName)
	{
		machineNameLen = strlen(machineName);
		machineNameW = (wchar_t *)alloca((machineNameLen + 1) * sizeof(wchar_t));
		mbstowcs(machineNameW, machineName, machineNameLen);
		machineNameW[machineNameLen] = 0;
	}

	if (domainName)
	{
		domainNameLen = strlen(domainName);
		domainNameW = (wchar_t *)alloca((domainNameLen + 1) * sizeof(wchar_t));
		mbstowcs(domainNameW, domainName, domainNameLen);
		domainNameW[domainNameLen] = 0;
	}

	status = NetGetDCName(machineNameW, domainNameW, &buffer);
	if (status == NERR_Success) {
		size_t pdcMachineNameLen = wcslen((wchar_t *)buffer);

		wcstombs(pdcMachineName, (wchar_t *)buffer, pdcMachineNameLen);
		pdcMachineName[pdcMachineNameLen] = 0;

		return 0;
	} else
		return -1;
}

//	enum trusted domains in machineName
extern "C" int __declspec(dllexport) EnumTrustedDomains(const char *machineName, char *domains)
{
	size_t machineNameLen;
	wchar_t *machineNameW;

	if (machineName)
	{
		machineNameLen = strlen(machineName);
		machineNameW = (wchar_t *)malloc((machineNameLen + 1) * sizeof(wchar_t));
		mbstowcs(machineNameW, machineName, machineNameLen);
		machineNameW[machineNameLen] = 0;
	}
	else
	{
		machineNameLen = 0;
		machineNameW = L"";
	}

	LSA_HANDLE lsaHandle;
	LSA_UNICODE_STRING lsaMachineName;

	//	weird lsa strings
	lsaMachineName.Length = machineNameLen * sizeof(wchar_t);
	lsaMachineName.MaximumLength = (machineNameLen + 1) * sizeof(wchar_t);
	lsaMachineName.Buffer = machineNameW;

	LSA_OBJECT_ATTRIBUTES lsaObjectAttributes;
	DWORD lsaAccessMask;

	ZeroMemory(&lsaObjectAttributes, sizeof(lsaObjectAttributes));
	lsaAccessMask = POLICY_VIEW_LOCAL_INFORMATION;

	if (LsaOpenPolicy(&lsaMachineName, &lsaObjectAttributes, lsaAccessMask, &lsaHandle))
		return -1;

	LSA_ENUMERATION_HANDLE lsaEnumHandle=0; // start an enum
	PLSA_TRUST_INFORMATION TrustInfo;
	ULONG ulReturned;               // number of items returned
	ULONG ulCounter;                // counter for items returned
	NTSTATUS Status;
	bool started = false;
	_bstr_t domainList("");
	int returnValue = 0;

	do {
		Status = LsaEnumerateTrustedDomains(
			lsaHandle,   // open policy handle
			&lsaEnumHandle, // enumeration tracker
			(void **)&TrustInfo,     // buffer to receive data
			32000,          // recommended buffer size
			&ulReturned     // number of items returned
			);
		//
		// get out if an error occurred
		//

		if( (Status != 0x00000000L) &&		//	success
			(Status != 0x00000105L) &&		//	more entries
			(Status != 0x8000001AL)		//	no more entries
			) {
				returnValue = -1;
				break;
			}

		//
		// Note: Sids are in TrustInfo[ulCounter].Sid
		//
		for(ulCounter = 0 ; ulCounter < ulReturned ; ulCounter++)
		{
			if (started)
				domainList += ";";

			started = true;

			domainList += (wchar_t *)(TrustInfo[ulCounter].Name.Buffer);

			returnValue++;		//	one more domain
		}

		//
		// free the buffer
		//
		LsaFreeMemory(TrustInfo);

	} while (Status != 0x8000001AL);		//	no more entries

	LsaClose(lsaHandle);

	if (returnValue && domainList.length())
		strcpy(domains, domainList);
	else
		domains[0] = 0;
		


	return returnValue;
}

extern "C" int __declspec(dllexport) EnumAdminAccounts(const char *machineName, char *accounts)
{
	USER_INFO_1 *userInfo = 0;
	NET_API_STATUS status;
	DWORD totalcount=0, resume=0, remaining=0, privileges, admincount=0;
	_bstr_t tmp("");
	wchar_t wuname[128];
	bool started = false;
	size_t machineNameLen;
	wchar_t *machineNameW;

	if (machineName)
		{
			machineNameLen = strlen(machineName);
			machineNameW = (wchar_t *)malloc((machineNameLen + 1) * sizeof(wchar_t));
			mbstowcs(machineNameW, machineName, machineNameLen);
			machineNameW[machineNameLen] = 0;
		}
		else
			machineNameW = 0;

	status = NetUserEnum(machineNameW, 1,
		FILTER_NORMAL_ACCOUNT|FILTER_INTERDOMAIN_TRUST_ACCOUNT|
		FILTER_WORKSTATION_TRUST_ACCOUNT|FILTER_SERVER_TRUST_ACCOUNT,
		(LPBYTE *)&userInfo, 64, &totalcount, &remaining, &resume );
	do
	{
		USER_INFO_1 *userInfoIt = userInfo;

		for (DWORD i=0;i<totalcount;i++,userInfoIt++)
			{
			privileges = userInfoIt->usri1_priv;
			if (privileges & USER_PRIV_ADMIN)
				{
				wcscpy( wuname, userInfoIt->usri1_name );
				if (started)
					tmp = tmp + ";";
				started = true;
				tmp = tmp + wuname;
				admincount++;
				}
			}

		if (userInfo)
			{
			NetApiBufferFree( userInfo );
			userInfo = 0;
			}

		if ( status == ERROR_MORE_DATA )
			{
			status = NetUserEnum(machineNameW, 1,
				FILTER_NORMAL_ACCOUNT|FILTER_INTERDOMAIN_TRUST_ACCOUNT|
				FILTER_WORKSTATION_TRUST_ACCOUNT|FILTER_SERVER_TRUST_ACCOUNT,
				(LPBYTE *)&userInfo, 64, &totalcount, &remaining, &resume );
			resume += totalcount;
			}
		else
			totalcount = 0;

	} while ( totalcount > 0 );

	strcpy( accounts, tmp );

	if (userInfo)
		NetApiBufferFree( userInfo );

	if (machineNameW)
		free(machineNameW);

	return admincount;
}

extern "C" long __declspec(dllexport) SetServiceRunAs(const char *service, const char *accounts, const char *pass)
{
	SC_HANDLE scMgr, scService;

	scMgr = OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );
	if (scMgr != NULL)
		{
		scService = OpenService( scMgr, service, SERVICE_ALL_ACCESS );
		if (scService != NULL)
			{
			if (!ChangeServiceConfig( scService,		// identifies the service
									SERVICE_NO_CHANGE,	// not changing service type
									SERVICE_NO_CHANGE,	// not changing start type
									SERVICE_NO_CHANGE,	// not changing error control
									NULL,					// not changing service binary file
									NULL,					// not changing load order
									NULL,				// not changing tag
									NULL,				// not changing dependencies
									accounts,			// account name - COMPUTERNAME\ID
									pass,				// account's password
									NULL ))				// not changing display name
				{
				return GetLastError();
				}
			CloseServiceHandle( scService );
			}
		else
			{
			return GetLastError();
			}
		CloseServiceHandle( scMgr );
		}
	else
		{
		return GetLastError();
		}
	return 0;		//success
}

extern "C" int __declspec(dllexport) SetPolicy( const char *acct )
{
	_bstr_t wacct( acct );
	LSA_HANDLE PolicyHandle;
	PSID pSid = 0;
	NTSTATUS Status;
	int iRetVal=0;

	Status = OpenPolicy( POLICY_CREATE_ACCOUNT | POLICY_LOOKUP_NAMES, &PolicyHandle );

	if(Status != STATUS_SUCCESS)
	{
#ifndef	NDEBUG
		char formatBuffer[4096];
		FormatLastError(GetLastError(), formatBuffer, "OpenPolicy()");
		MessageBox(0, formatBuffer, "ERROR!!!!!", MB_OK);
#endif

		return 0;
	}

	// note: wacct is in format of computername\username
	if(GetPrincipalSID( wacct, &pSid ) == ERROR_SUCCESS)
	{
		// Grant the SeServiceLogonRight to users represented by pSid.
		Status = SetPrivilegeOnAccount( PolicyHandle, pSid, L"SeServiceLogonRight" );

		if(Status == STATUS_SUCCESS)
		{
			iRetVal = 1;
			// Grant the SeBatchLogonRight to users represented by pSid.
			Status = SetPrivilegeOnAccount( PolicyHandle, pSid, L"SeBatchLogonRight" );
			if (Status !=STATUS_SUCCESS)
			{
#ifndef	NDEBUG
				char formatBuffer[4096];
				FormatLastError(GetLastError(), formatBuffer, "SetPrivilegeOnAccount(SeBatchLogonRight)");
				MessageBox(0, formatBuffer, "ERROR!!!!!", MB_OK);
#endif

				iRetVal = 0;
			}
		}
		else
		{
#ifndef	NDEBUG
			char formatBuffer[4096];
			FormatLastError(GetLastError(), formatBuffer, "SetPrivilegeOnAccount(SeServiceLogonRight)");
			MessageBox(0, formatBuffer, "ERROR!!!!!", MB_OK);
#endif

			iRetVal = 0;
		}
	}
	else
	{
#ifndef	NDEBUG
		char formatBuffer[4096];
		FormatLastError(GetLastError(), formatBuffer, "GetPrincipalSID()");
		MessageBox(0, formatBuffer, "ERROR!!!!!", MB_OK);
#endif

		iRetVal = 0;
	}

	// Close the policy handle.
	LsaClose(PolicyHandle);

	// Free memory allocated for SID.
	if (pSid != NULL)
		HeapFree(GetProcessHeap(), 0, pSid);

	return iRetVal;
}

extern "C" int __declspec(dllexport) LaunchMinimized( const char *file, char *parms, const char *path )
{
	PROCESS_INFORMATION pInfo;
	char cmdBuf[512];
	STARTUPINFO startInfo;
	DWORD retCode;

	GetStartupInfo(&startInfo);
	startInfo.dwFlags = STARTF_USESHOWWINDOW;
	startInfo.wShowWindow = SW_SHOWMINIMIZED;
	// note 'file' is a fully qualifed file name (including extension)
	sprintf( cmdBuf, "%s %s", file, parms );

	if ( !CreateProcess( 0, cmdBuf, 0, 0, FALSE, 0, 0, path, &startInfo, &pInfo ) )
		return GetLastError();


	// wait for it to finish....
	if ( WaitForSingleObject( pInfo.hProcess, INFINITE ) == WAIT_FAILED )
		return GetLastError();

	GetExitCodeProcess( pInfo.hProcess, &retCode );

	return 	retCode;
	
}

extern "C" int __declspec(dllexport) FlushRegistryKey( const char *keyname )
{
	HKEY	key;
	long	ret;

	ret = RegOpenKeyEx( HKEY_LOCAL_MACHINE, keyname, 0, KEY_ALL_ACCESS, &key );
	if(!key) 
		return ret;

	ret = RegFlushKey( key );

	if (ret != ERROR_SUCCESS)
		return ret;

	RegCloseKey( key );

	return ERROR_SUCCESS;
}

extern "C" long __declspec(dllexport) RemoveService(const char *service)
{
#define ERROR_SERVICE_FAILED_TO_STOP     000010001
	SC_HANDLE scMgr, scService;
	int cWait = 0;

	scMgr = OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );
	if (scMgr != NULL)
		{
		scService = OpenService( scMgr, service, SERVICE_ALL_ACCESS );
		if (scService != NULL)
			{
			// remove the service
			if ( !DeleteService( scService ) )
				return GetLastError();
			CloseServiceHandle( scService );
			return 0;
			}
		else
			{
			return GetLastError();
			}
		CloseServiceHandle( scMgr );
		}
	else
		{
		return GetLastError();
		}
	return 0;		//success
}

extern "C" long __declspec(dllexport) StopService(const char *service)
{
#define ERROR_SERVICE_FAILED_TO_STOP     000010001
	SC_HANDLE scMgr, scService;
	SERVICE_STATUS sStatus;
	int cWait = 0;

	scMgr = OpenSCManager( NULL, NULL, SC_MANAGER_ALL_ACCESS );
	if (scMgr != NULL)
		{
		scService = OpenService( scMgr, service, SERVICE_ALL_ACCESS );
		if (scService != NULL)
			{
			if ( !ControlService( scService, SERVICE_CONTROL_STOP, &sStatus ) )
				{
				CloseServiceHandle( scService );
				return GetLastError();
				}
			else
				{
				while( QueryServiceStatus( scService, &sStatus ) )
					{
					if ( cWait > 240 )
						{
						CloseServiceHandle( scService );
						return ERROR_SERVICE_FAILED_TO_STOP;
						}
					if ( sStatus.dwCurrentState == SERVICE_STOP_PENDING )
						{
						Sleep( 250 );
						cWait++;
						}
					else if( sStatus.dwCurrentState == SERVICE_STOPPED )
						{
						CloseServiceHandle( scService );
						return 0;
						}
					}
				if ( sStatus.dwCurrentState != SERVICE_STOPPED )
					{
					CloseServiceHandle( scService );
					return GetLastError();
					}
				CloseServiceHandle( scService );
				return 0;
				}

			CloseServiceHandle( scService );
			}
		else
			{
			return GetLastError();
			}
		CloseServiceHandle( scMgr );
		}
	else
		{
		return GetLastError();
		}
	return 0;		//success
}

extern "C" DWORD __declspec(dllexport) ChangeAppIDAccessACL( LPTSTR AppID, LPTSTR Principal, BOOL SetPrincipal, BOOL Permit )
{

    TCHAR   keyName [256];

	if (AppID [0] == '{')
        wsprintf (keyName, TEXT("APPID\\%s"), AppID);
	else
        wsprintf (keyName, TEXT("APPID\\{%s}"), AppID);

    if (SetPrincipal)
    {
        RemovePrincipalFromNamedValueSD (HKEY_CLASSES_ROOT, _bstr_t(keyName), L"AccessPermission", _bstr_t(Principal));
        return AddPrincipalToNamedValueSD (HKEY_CLASSES_ROOT, _bstr_t(keyName), L"AccessPermission", _bstr_t(Principal), Permit);
    } else
        return RemovePrincipalFromNamedValueSD (HKEY_CLASSES_ROOT, _bstr_t(keyName), L"AccessPermission", _bstr_t(Principal));
}

extern "C" DWORD __declspec(dllexport) ChangeAppIDLaunchACL( LPTSTR AppID, LPTSTR Principal, BOOL SetPrincipal, BOOL Permit )
{
    TCHAR   keyName [256];

	if (AppID [0] == '{')
        wsprintf (keyName, TEXT("APPID\\%s"), AppID);
	else
        wsprintf (keyName, TEXT("APPID\\{%s}"), AppID);

    if (SetPrincipal)
    {
        RemovePrincipalFromNamedValueSD (HKEY_CLASSES_ROOT, _bstr_t(keyName), L"LaunchPermission", _bstr_t(Principal));
        return AddPrincipalToNamedValueSD (HKEY_CLASSES_ROOT, _bstr_t(keyName), L"LaunchPermission", _bstr_t(Principal), Permit);
    } else
        return RemovePrincipalFromNamedValueSD (HKEY_CLASSES_ROOT, _bstr_t(keyName), L"LaunchPermission", _bstr_t(Principal));

}


extern "C" DWORD __declspec(dllexport) GetRunAsPassword( LPTSTR AppID, LPTSTR Password )
{
    LSA_OBJECT_ATTRIBUTES objectAttributes;
    HANDLE                policyHandle = NULL;
    LSA_UNICODE_STRING    lsaKeyString;
    PLSA_UNICODE_STRING   lsaPasswordString;
    WCHAR                 key [4 + GUIDSTR_MAX + 1];
    WCHAR                 wideAppID [GUIDSTR_MAX + 1];
    ULONG                 returnValue;

#ifndef UNICODE
    STR2UNI (wideAppID, AppID);
#else
    lstrcpy (wideAppID, AppID);
#endif

    wcscpy (key, L"SCM:");
    wcscat (key, wideAppID);

    lsaKeyString.Length = (USHORT) ((wcslen (key) + 1) * sizeof (WCHAR));
    lsaKeyString.MaximumLength = (GUIDSTR_MAX + 5) * sizeof (WCHAR);
    lsaKeyString.Buffer = key;

    //
    // Open the local security policy
    //

    memset (&objectAttributes, 0x00, sizeof (LSA_OBJECT_ATTRIBUTES));
    objectAttributes.Length = sizeof (LSA_OBJECT_ATTRIBUTES);

    returnValue = LsaOpenPolicy (NULL,
                                 &objectAttributes,
                                 POLICY_GET_PRIVATE_INFORMATION,
                                 &policyHandle);

    if (returnValue != ERROR_SUCCESS)
        return returnValue;

    //
    // Read the user's password
    //

    returnValue = LsaRetrievePrivateData (policyHandle,
                                          &lsaKeyString,
                                          &lsaPasswordString);

    if (returnValue != ERROR_SUCCESS)
    {
        LsaClose (policyHandle);
        return returnValue;
    }

    LsaClose (policyHandle);

#ifndef UNICODE
    UNI2STR (Password, lsaPasswordString->Buffer);
#else
    wcscpy (Password, lsaPasswordString->Buffer);
#endif

    return ERROR_SUCCESS;
}

HRESULT
SetRAP (
	LPWSTR AppID,
	LPWSTR Principal,
	LPWSTR Password
	)
{
	LSA_OBJECT_ATTRIBUTES objectAttributes;
	HANDLE                policyHandle = NULL;
	LSA_UNICODE_STRING    lsaKeyString;
	LSA_UNICODE_STRING    lsaPasswordString;
	WCHAR                 key [4 + GUIDSTR_MAX + 1];
	DWORD                 returnValue;
	WCHAR                 keyName [MAX_PATH+1];
	HKEY                  registryKey;
	DWORD                 valueType;
	DWORD                 valueSize = 15;

	swprintf (keyName, L"AppID\\%s", AppID);
	returnValue = RegOpenKeyExW (HKEY_CLASSES_ROOT, keyName, 0, KEY_ALL_ACCESS, &registryKey);
	if (returnValue == ERROR_SUCCESS)
	{
		returnValue = RegQueryValueExW (registryKey, L"LocalService", NULL, &valueType, NULL, &valueSize);
		RegCloseKey (registryKey);

	if (returnValue == ERROR_SUCCESS || returnValue == ERROR_MORE_DATA)
		return DCP_E_OBJ_IS_A_SERVICE;
	}
	else
	{
		ReportFailure( "SetRAP", "RegOpenKeyEx", __FILE__, __LINE__ );
	}

	//
	// Enable the batch right if a _real_ user was specified
	//

	if (wcsicmp (Principal, L"Interactive User") != 0)
	{
		returnValue = SetAccountRights (Principal, L"SeBatchLogonRight");
		if (returnValue != ERROR_SUCCESS)
		{
			if (returnValue == HRESULT_FROM_WIN32 (ERROR_NONE_MAPPED))
				return DCP_E_UNKNOWN_USER; else
			{
				ReportFailure("SetRAP", "SetAccountRights", __FILE__, __LINE__);
				return HRESULT_FROM_WIN32 (returnValue);
			}
		}
	}

	wcscpy (key, L"SCM:");
	wcscat (key, AppID);

	lsaKeyString.Length = (USHORT) ((wcslen (key)+1) * sizeof (WCHAR));
	lsaKeyString.MaximumLength = lsaKeyString.Length + sizeof (WCHAR);
	lsaKeyString.Buffer = key;

	lsaPasswordString.Length = (USHORT) ((wcslen (Password)+1) * sizeof (WCHAR));
	lsaPasswordString.Buffer = Password;
	lsaPasswordString.MaximumLength = lsaPasswordString.Length + sizeof (WCHAR);

    //
    // Open the local security policy
    //

	memset (&objectAttributes, 0x00, sizeof (LSA_OBJECT_ATTRIBUTES));
	objectAttributes.Length = sizeof (LSA_OBJECT_ATTRIBUTES);

	returnValue = LsaOpenPolicy (NULL,
								&objectAttributes,
								POLICY_CREATE_SECRET,
								&policyHandle);

	if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("SetRAP", "LsaOpenPolicy", __FILE__, __LINE__);
		return HRESULT_FROM_WIN32 (returnValue);
	}

	//
	// Store the user's password
	//

	returnValue = LsaStorePrivateData (policyHandle,
										&lsaKeyString,
										&lsaPasswordString);

	if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("SetRAP", "LsaStorePrivateData", __FILE__, __LINE__);
		LsaClose (policyHandle);
		return HRESULT_FROM_WIN32 (returnValue);
	}

	LsaClose (policyHandle);

	return ERROR_SUCCESS;
}

extern "C" DWORD __declspec(dllexport) SetRunAsPassword( LPTSTR AppID, LPTSTR Principal, LPTSTR Password )
{
	LPWSTR appid;
	WCHAR keyName [256];
	DWORD returnValue;
	HKEY registryKey;

	swprintf (keyName, L"APPID\\%s", (wchar_t *)(_bstr_t(AppID)));

	returnValue = RegOpenKeyExW (HKEY_CLASSES_ROOT, keyName, 0, KEY_ALL_ACCESS, &registryKey);

	if (returnValue != ERROR_SUCCESS)
	{
		if (returnValue == ERROR_FILE_NOT_FOUND)
			return DCP_E_INVALID_OBJECT; else
			return HRESULT_FROM_WIN32 (returnValue);
	}

	returnValue = RegSetValueExW (registryKey, L"RunAs", 0, REG_SZ, (LPBYTE)((wchar_t *)(_bstr_t(Principal))), strlen (Principal) * sizeof (WCHAR));

	if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("SetRunAsPassword", "RegSetValueExW", __FILE__, __LINE__);
		return HRESULT_FROM_WIN32 (returnValue);
	}

	RegCloseKey (registryKey);

	wcstok (keyName, L"\\");
	appid = wcstok (NULL, L"\\");

	returnValue = SetRAP (appid, _bstr_t(Principal), _bstr_t(Password));

	if (returnValue != ERROR_SUCCESS)
	{
		ReportFailure("SetRunAsPassword", "SetRAP", __FILE__, __LINE__);
		return returnValue;
	}

	return S_OK;
}